/***********************************************************

    File:       tclstdio.c

    Contents:   C stdio redirection to Tcl procedures

    Author:     Dr. Rico Cozzio,

                COMEXAR Engineering AG
                Breitenstr. 15
                Postfach 441
                CH-8853 Lachen
                Switzerland

    E-mail:     cozzio@ln.active.ch
    WWW:        http://www2.active.ch/~cozzio/index.html

    History:    22.12.1995  First version
                10.02.1996  Separate functions for writing
                            to stdout and stderr
                18.07.1996  Cleaned up for distribution

    Usage:      Include the file tclstdio.h instead of stdio.h
                in every C file that uses stdio and define
                the symbol TCLSTDIO.
                Additionally, the following Tcl procedures
                must be defined:
                    stdio_puts_stdout   string
                    stdio_puts_stderr   string
                    stdio_read          bytes

**************** LEGAL STUFF ********************************

    Copyright (c) COMEXAR Engineering AG, 1996

    License is granted to copy, to use, and to make and to use
    derivative works for any purpose, provided that this
    copyright notice and this license notice is included in all
    copies and any derivatives works.
    The licensee acknowleges that Rico Cozzio and COMEXAR
    Corporation have no liability for licensee's use
    or for any derivative works  made by licensee.
    The COMEXAR name shall not be used in any advertising or the
    like without its written permission.

    This software is provided AS IS. COMEXAR CORPORATION
    DISCLAIMS AND LICENSEE AGREES THAT ALL WARRANTIES, EXPRESS
    OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
    PURPOSE. NOTWITHSTANDING ANY OTHER PROVISION CONTAINED
    HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM THE
    SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, INCLUDING
    CONSEQUENTIAL OR ANY OTHER INDIRECT DAMAGES, WHETHER ARISING
    IN CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY,
    EVEN IF COMEXAR CORPORATION IS ADVISED OF THE POSSIBILITY OF
    SUCH DAMAGES."

    This code was adapted from the TkNT distribution by
    Gordon Chaffee, see acknowledgement below.
    However, almost none of his original code survived.

*************************************************************

    winstdio.c

    Copyright (c) 1994 Software Research Associates, Inc.

    Permission to use, copy, modify, and distribute this software and its
    documentation for any purpose and without fee is hereby granted, provided
    that the above copyright notice appear in all copies and that both that
    copyright notice and this permission notice appear in supporting
    documentation, and that the name of Software Research Associates not be
    used in advertising or publicity pertaining to distribution of the
    software without specific, written prior permission.  Software Research
    Associates makes no representations about the suitability of this software
    for any purpose.  It is provided "as is" without express or implied
    warranty.

*************** END OF LEGAL STUFF *************************/

#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <tcl.h>
#include "tclstdio.h"

// This is only really required for Win32
#ifndef __WIN32__
#   if defined(_WIN32) || defined(WIN32)
#	define __WIN32__
#   endif
#endif

#if defined(__WIN32__)
	#   define WIN32_LEAN_AND_MEAN
	#   include <windows.h>
	#   undef WIN32_LEAN_AND_MEAN
	#   if defined(_MSC_VER)
		#	define EXPORT(a,b) __declspec(dllexport) a b
		#	define DllEntryPoint DllMain
	#   else
		#	if defined(__BORLANDC__)
		#	    define EXPORT(a,b) a _export b
		#	else
		#	    define EXPORT(a,b) a b
		#	endif
	#   endif
	BOOL APIENTRY \
	DllEntryPoint(HINSTANCE hInst, DWORD reason, LPVOID reserved) { \
		return TRUE; \
	}

#endif

extern int     fileno(FILE *fd);

static Tcl_Interp* tclInterp;

#ifndef TCLSTDIO_LIBRARY
#define TCLSTDIO_LIBRARY ""
#endif
static char defaultLibraryDir[200] = TCLSTDIO_LIBRARY;

int
Tclstdio_Init( Tcl_Interp *interp )
{
    static char initScript[] =
    "set stdiodirs {}\n\
    if [info exists env(TCLSTDIO_LIBRARY)] {\n\
	lappend stdiodirs $env(TCLSTDIO_LIBRARY)\n\
    }\n\
    if [info exists env(EXT_FOLDER)] {\n\
	lappend stdiodirs [file join $env(EXT_FOLDER) \"Tool Command Language\" lib tclstdio$tclstdioversion]\n\
    }\n\
    lappend stdiodirs $stdiolibrary\n\
    set stdiolibrary {}\n\
    if ![catch {uplevel #0 source -rsrc tclstdio}] {\n\
	return\n\
    }\n\
    lappend stdiodirs [file join [file dirname [info library]] tclstdio$tclstdioversion]\n\
    set parentDir [file dirname [file dirname [info nameofexecutable]]]\n\
    lappend stdiodirs [file join $parentDir lib tclstdio$tclstdioversion]\n\
    lappend stdiodirs [file join [file dirname $parentDir] tclstdio library]\n\
    lappend stdiodirs [file join $parentDir library]\n\
    foreach i $stdiodirs {\n\
	set stdiolibrary $i\n\
	if ![catch {uplevel #0 source \\{[file join $i tclstdio.tcl]\\}}] {\n\
	    break\n\
	}\n\
	set stdiolibrary {}\n\
    }\n\
    if {$stdiolibrary == {}} {\n\
    set msg \"Can't find a usable tclstdio.tcl in the following directories: \n\"\n\
    append msg \"    $stdiodirs\n\"\n\
    append msg \"This probably means that tclstdio wasn't installed properly\n\"\n\
    append msg \"or you need to set your TCLSTDIO_LIBRARY environment variable?\"\n\
    error $msg\n\
    }\n\
    set tclstdioversion\n\
    ";

	char* libDir;
	
    Tcl_SetVar(interp, "tclstdioversion", "0.1", TCL_GLOBAL_ONLY);	
    libDir = Tcl_GetVar(interp, "stdiolibrary", TCL_GLOBAL_ONLY);
    if (libDir == NULL) {
		Tcl_SetVar(interp, "stdiolibrary", defaultLibraryDir, TCL_GLOBAL_ONLY);
    }
	
    Tcl_PkgProvide(interp,"Tclstdio","0.1");

	tclInterp = interp;
    return Tcl_Eval(interp, initScript);
}

#define BUFFER_LENGTH 5120

/***********************************************************
    Output to Tcl stdout & stderr
************************************************************/

int tclputs_stdout(char *string)
{
/*
    Tcl_CmdInfo info;
    char *argv[] = {"puts","-nonewline","stdout",string,NULL};

    if (Tcl_GetCommandInfo(tclInterp,"puts",&info) &&
        (*info.proc)(info.clientData,tclInterp,4,argv) == TCL_OK) {
*/
    Tcl_DString cmd;
    Tcl_DStringInit(&cmd);
    Tcl_DStringAppend(&cmd,"stdio_puts_stdout",-1);
    Tcl_DStringAppendElement(&cmd,string);
    if (Tcl_Eval(tclInterp,Tcl_DStringValue(&cmd)) == TCL_OK) {
        Tcl_DStringFree(&cmd);
        return 1;
    } else {
        Tcl_DStringFree(&cmd);
        return EOF;
    }
}

int tclputs_stderr(char *string)
{
/*
    Tcl_CmdInfo info;
    char *argv[] = {"puts","-nonewline","stderr",string,NULL};

    if (Tcl_GetCommandInfo(tclInterp,"puts",&info) &&
        (*info.proc)(info.clientData,tclInterp,4,argv) == TCL_OK) {
*/
    Tcl_DString cmd;
    Tcl_DStringInit(&cmd);
    Tcl_DStringAppend(&cmd,"stdio_puts_stderr",-1);
    Tcl_DStringAppendElement(&cmd,string);
    if (Tcl_Eval(tclInterp,Tcl_DStringValue(&cmd)) == TCL_OK) {
        Tcl_DStringFree(&cmd);
        return 1;
    } else {
        Tcl_DStringFree(&cmd);
        return EOF;
    }
}

/***********************************************************
    Input from Tcl stdin
************************************************************/

char *tclgets(char *buffer,int max)
{
/*
    Tcl_CmdInfo info;
    char *argv[] = {"read","stdin",maxnum,NULL};

    sprintf(maxnum,"%d",max);
    if (Tcl_GetCommandInfo(tclInterp,"read",&info) &&
        (*info.proc)(info.clientData,tclInterp,3,argv) == TCL_OK) {
*/
    char maxnum[20] = "";
    Tcl_DString cmd;
    Tcl_DStringInit(&cmd);
    Tcl_DStringAppend(&cmd,"stdio_read",-1);
    sprintf(maxnum,"%d",max);
    Tcl_DStringAppendElement(&cmd,maxnum);
    if (Tcl_Eval(tclInterp,Tcl_DStringValue(&cmd)) == TCL_OK) {
        strncpy(buffer,tclInterp->result,max);
        Tcl_DStringFree(&cmd);
        return buffer;
    } else {
        Tcl_DStringFree(&cmd);
        return NULL;
    }
}

/***********************************************************
    Stdio replacement functions
************************************************************/

int
tclio_printf(char * fmt, ...)
{
    va_list     arg_ptr;
    int result;

    va_start(arg_ptr, fmt);
    {   char buffer[BUFFER_LENGTH + 1];
            if((result = vsprintf(buffer, fmt, arg_ptr)) > 0){
                buffer[BUFFER_LENGTH] = '\0';
            tclputs_stdout(buffer);
        }
    }
    va_end(arg_ptr);
    return result;
}

int
tclio_fprintf(FILE *fp, char *fmt, ...)
{
    va_list     arg_ptr;
    int result;

    va_start(arg_ptr, fmt);
    {   char buffer[BUFFER_LENGTH + 1];
            if (fp == stdin) {
                result = -1;
            } else if(fp == stdout || fp == stderr) {
                result = vsprintf(buffer, fmt, arg_ptr);
            if (fp == stdout) {
                tclputs_stdout(buffer);
            } else {
                tclputs_stderr(buffer);
            }
            } else {
                result = vfprintf(fp, fmt, arg_ptr);
            }
    }
    va_end(arg_ptr);
    return result;
}

int
tclio_fputc(int ch, FILE *f)
{
    char c;
	char s[2];
	
	if (f == stdout || f == stderr) {
		s[0] = (char)ch;
		s[1] = '\0';
        if (f == stdout) {
            tclputs_stdout(s);
        } else {
            tclputs_stderr(s);
        }
            return ch;
    } else {
            return fputc(ch, f);
    }
}

int
tclio_fputs(const char *s, FILE *f)
{
        if (f == stdout || f == stderr) {
        if (f == stdout) {
            tclputs_stdout((char*)s);
        } else {
            tclputs_stderr((char*)s);
        }
            return 1;
    } else {
            return fputs(s, f);
    }
}

void
tclio_clearerr(FILE *fp)
{
        if (fp == stdin || fp == stdout || fp == stderr) {
            return;
        } else {
        clearerr(fp);
    }
}

int
tclio_fileno(FILE *fp)
{
	if (fp == stdin) {
            return 0;
        } else if (fp == stdout) {
            return 1;
        } else if (fp == stderr) {
            return 2;
    } else {
        return fileno(fp);
    }
}

int
tclio_ferror(FILE *fp)
{
        if (fp == stdin || fp == stdout || fp == stderr) {
            return 0;
    } else {
        return ferror(fp);
    }
}

int
tclio_feof(FILE *fp)
{
        if (fp == stdin || fp == stdout || fp == stderr) {
            return 0;
    } else {
        return feof(fp);
    }
}

int
tclio_fflush(FILE *fp)
{
        if (fp == stdin || fp == stdout || fp == stderr) {
            return 0;
    } else {
        return fflush(fp);
    }
}

char *
tclio_fgets(char *buffer, int length, FILE *fp)
{
    if (fp == stdin) {
        return tclgets(buffer,length);
    } else {
        return fgets(buffer, length, fp);
    }
}

static int stdinchar='\0',stdinflag=0;

int
tclio_fgetc(FILE *f)
{
    char c;

        if (f == stdin) {
            if (stdinflag) {
                    stdinflag=0;
                    return stdinchar;
            }
        {
            char buffer[2] = {'\0','\0'};
            while(tclgets(buffer,1) == NULL || buffer[0] == '\0') {
            }
            return buffer[0];
        }
        } else {
            return fgetc(f);
    }
}

int
tclio_fungetc(int c,FILE *f)
{
        if (f == stdin) {
                 stdinchar=c;
                 stdinflag=1;
                 return 0;
    } else {
            return ungetc(c,f);
    }
}

/*
char* tclio_tmpnam(char *) {
}
*/